/**
 * Copyright (c) 2009 Mark Logic Corporation. All rights reserved.
 */
package com.marklogic.cpox;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

/**
 * @author Michael Blakeley, Mark Logic Corporation
 * 
 */
public class Timer {

    public static final int BYTES_PER_KILOBYTE = 1024;

    public static final int MILLISECONDS_PER_SECOND = 1000;

    public static final int NANOSECONDS_PER_MICROSECOND = MILLISECONDS_PER_SECOND;

    public static final int MICROSECONDS_PER_MILLISECOND = MILLISECONDS_PER_SECOND;

    public static final int NANOSECONDS_PER_MILLISECOND = NANOSECONDS_PER_MICROSECOND
            * MICROSECONDS_PER_MILLISECOND;

    public static final int NANOSECONDS_PER_SECOND = NANOSECONDS_PER_MILLISECOND
            * MILLISECONDS_PER_SECOND;

    protected long start;

    protected long eventCount;

    protected long errors = 0;

    protected long bytes = 0;

    protected long duration = -1;

    protected ArrayList<TimedEvent> events = new ArrayList<TimedEvent>();

    public Timer() {
        start = System.nanoTime();
        eventCount = 0;
    }

    /**
     * @param _count
     * @param _duration
     */
    public Timer(long _count, long _duration) {
        start = System.nanoTime();
        duration = _duration;
        eventCount = _count;
    }

    public void add(TimedEvent event) {
        add(event, true);
    }

    public void add(TimedEvent event, boolean keepEvent) {
        // in case the user forgot to call stop(): note that bytes won't be
        // counted!
        event.stop();
        synchronized (events) {
            bytes += event.getBytes();
            if (event.isError()) {
                errors++;
            }
            eventCount++;
            // to reduce memory utilization, keeping the event is optional
            if (keepEvent) {
                events.add(event);
            }
        }
    }

    /**
     * @param _timer
     */
    public void add(Timer _timer) {
        _timer.stop();
        synchronized (events) {
            bytes += _timer.getBytes();
            errors += _timer.getErrorCount();
            events.addAll(_timer.events);
            eventCount += _timer.eventCount;
        }
    }

    /**
     * @return
     */
    public long getBytes() {
        return bytes;
    }

    /**
     * @return
     */
    public long getEventCount() {
        return eventCount;
    }

    /**
     * @return
     */
    public long getSuccessfulEventCount() {
        return eventCount - errors;
    }

    /**
     * @return
     */
    public long getErrorCount() {
        return errors;
    }

    /**
     * @return
     */
    public long getDuration() {
        if (duration < 0)
            return (System.nanoTime() - start);

        return duration;
    }

    /**
     * @return
     */
    public long getMeanOfEvents() {
        if (eventCount < 1)
            return 0;

        long sum = 0;
        for (int i = 0; i < eventCount; i++) {
            sum += events.get(i).getDuration();
        }

        return Math.round((double) sum / eventCount);
    }

    /**
     * @return
     */
    public long getPercentileDuration(int p) {
        if (eventCount < 1)
            return 0;

        double size = eventCount;
        Comparator<TimedEvent> c = new TimedEventDurationComparator();
        Collections.sort(events, c);
        int pidx = (int) (p * size * .01);
        return events.get(pidx).getDuration();
    }

    /**
     * @return
     */
    public long getMaxDuration() {
        long max = 0;
        for (int i = 0; i < eventCount; i++)
            max = Math.max(max, events.get(i).getDuration());
        return max;
    }

    /**
     * @return
     */
    public long getMinDuration() {
        long min = Integer.MAX_VALUE;
        for (int i = 0; i < eventCount; i++)
            min = Math.min(min, events.get(i).getDuration());
        return min;
    }

    /**
     * @return
     */
    public double getMeanOverall() {
        return (double) getDuration() / (double) eventCount;
    }

    /**
     * @return
     */
    public double getMeanOverallMilliseconds() {
        return getDurationMilliseconds() / eventCount;
    }

    /**
     * @return
     */
    public long getStart() {
        return start;
    }

    /**
     * @return
     */
    public long getKiloBytes() {
        return (long) ((double) bytes / BYTES_PER_KILOBYTE);
    }

    public double getKilobytesPerSecond() {
        return ((double) bytes / BYTES_PER_KILOBYTE)
                / getDurationSeconds();
    }

    public double getEventsPerSecond() {
        // events per second
        return eventCount / getDurationSeconds();
    }

    /**
     * @param l
     */
    public long stop() {
        return stop(System.nanoTime());
    }

    /**
     * @param l
     */
    public synchronized long stop(long l) {
        if (duration < 0) {
            duration = l - start;
        }

        return duration;
    }

    /**
     * @param _duration
     */
    public void incrementDuration(long _duration) {
        duration += _duration;
    }

    /**
     * @param count
     * @param successful
     */
    public void incrementEventCount(int count, boolean successful) {
        eventCount += count;
        if (!successful) {
            errors++;
        }
    }

    /**
     * 
     */
    public void incrementEventCount() {
        incrementEventCount(1, true);
    }

    /**
     * @param count
     */
    public void incrementEventCount(int count) {
        incrementEventCount(count, true);
    }

    /**
     * @param successful
     */
    public void incrementEventCount(boolean successful) {
        incrementEventCount(1, successful);
    }

    /**
     * 
     */
    public void incrementErrorCount() {
        incrementEventCount(false);
    }

    /**
     * @param _count
     */
    public void setEventCount(long _count) {
        eventCount = _count;
    }

    /**
     * @return
     */
    public double getDurationMilliseconds() {
        // ns to seconds
        return ((double) getDuration())
                / ((double) NANOSECONDS_PER_MILLISECOND);
    }

    /**
     * @return
     */
    public double getDurationSeconds() {
        // ns to seconds
        return ((double) getDuration())
                / ((double) NANOSECONDS_PER_SECOND);
    }

    public String getProgressMessage(boolean rawValues) {
        return (rawValues ? eventCount + " events, " + getBytes()
                + " B in " + getDurationSeconds() + " s, " : "")
                + Math.round(getEventsPerSecond())
                + " events/s, "
                + Math.round(getKilobytesPerSecond()) + " kB/s";
    }

    /**
     * @return
     */
    public String getProgressMessage() {
        return getProgressMessage(false);
    }

}
